@@ -46,6 +46,7 @@ from dohq_artifactory.admin import RepositoryRemote
 from dohq_artifactory.admin import RepositoryVirtual
 from dohq_artifactory.admin import User
 from dohq_artifactory.auth import XJFrogArtApiAuth
+from dohq_artifactory.exception import ArtifactoryException
 
 try:
     import requests.packages.urllib3 as urllib3
@@ -244,6 +245,25 @@ def chunks(data, size):
         yield {k: data[k] for k in islice(it, size)}
 
 
+def log_download_progress(bytes_now, total_size):
+    """
+    Function to log download progress
+    :param bytes_now: current number of bytes
+    :param total_size: total file size in bytes
+    :return:
+    """
+    if total_size > 0:
+        msg = "Downloaded {}/{}MB...[{}%]".format(
+            int(bytes_now / 1024 / 1024),
+            int(total_size / 1024 / 1024),
+            round(bytes_now / total_size * 100, 2),
+        )
+    else:
+        msg = "Downloaded {}MB".format(int(bytes_now / 1024 / 1024))
+
+    logging.debug(msg)
+
+
 class HTTPResponseWrapper(object):
     """
     This class is intended as a workaround for 'requests' module
@@ -329,6 +349,29 @@ def encode_properties(parameters):
     return ";".join(result)
 
 
+# Declare contextlib class that was enabled in Py 3.7. Declare for compatibility with 3.5
+# this class is taken and modified from standard module contextlib
+class nullcontext:
+    """Context manager that does no additional processing.
+
+    Used as a stand-in for a normal context manager, when a particular
+    block of code is only sometimes used with a normal context manager:
+
+    cm = optional_cm if condition else nullcontext()
+    with cm:
+        # Perform operation, using optional_cm if condition is True
+    """
+
+    def __init__(self, enter_result=None):
+        self.enter_result = enter_result
+
+    def __enter__(self):
+        return self.enter_result
+
+    def __exit__(self, *excinfo):
+        pass
+
+
 class _ArtifactoryFlavour(pathlib._Flavour):
     """
     Implements Artifactory-specific pure path manipulations.
@@ -565,58 +608,113 @@ class _ArtifactoryAccessor(pathlib._Accessor):
     """
 
     def rest_get(
-        self, url, params=None, headers=None, session=None, verify=True, cert=None
+        self,
+        url,
+        params=None,
+        headers=None,
+        session=None,
+        verify=True,
+        cert=None,
+        timeout=None,
     ):
         """
         Perform a GET request to url with requests.session
         """
-        res = session.get(url, params=params, headers=headers, verify=verify, cert=cert)
+        res = session.get(
+            url,
+            params=params,
+            headers=headers,
+            verify=verify,
+            cert=cert,
+            timeout=timeout,
+        )
         return res.text, res.status_code
 
     def rest_put(
-        self, url, params=None, headers=None, session=None, verify=True, cert=None
+        self,
+        url,
+        params=None,
+        headers=None,
+        session=None,
+        verify=True,
+        cert=None,
+        timeout=None,
     ):
         """
         Perform a PUT request to url with requests.session
         """
-        res = session.put(url, params=params, headers=headers, verify=verify, cert=cert)
+        res = session.put(
+            url,
+            params=params,
+            headers=headers,
+            verify=verify,
+            cert=cert,
+            timeout=timeout,
+        )
         return res.text, res.status_code
 
     def rest_post(
-        self, url, params=None, headers=None, session=None, verify=True, cert=None
+        self,
+        url,
+        params=None,
+        headers=None,
+        session=None,
+        verify=True,
+        cert=None,
+        timeout=None,
     ):
         """
         Perform a POST request to url with requests.session
         """
         res = session.post(
-            url, params=params, headers=headers, verify=verify, cert=cert
+            url,
+            params=params,
+            headers=headers,
+            verify=verify,
+            cert=cert,
+            timeout=timeout,
         )
         return res.text, res.status_code
 
-    def rest_del(self, url, params=None, session=None, verify=True, cert=None):
+    def rest_del(
+        self, url, params=None, session=None, verify=True, cert=None, timeout=None
+    ):
         """
         Perform a DELETE request to url with requests.session
         """
-        res = session.delete(url, params=params, verify=verify, cert=cert)
+        res = session.delete(
+            url, params=params, verify=verify, cert=cert, timeout=timeout
+        )
         return res.text, res.status_code
 
     def rest_put_stream(
-        self, url, stream, headers=None, session=None, verify=True, cert=None
+        self,
+        url,
+        stream,
+        headers=None,
+        session=None,
+        verify=True,
+        cert=None,
+        timeout=None,
     ):
         """
         Perform a chunked PUT request to url with requests.session
         This is specifically to upload files.
         """
-        res = session.put(url, headers=headers, data=stream, verify=verify, cert=cert)
+        res = session.put(
+            url, headers=headers, data=stream, verify=verify, cert=cert, timeout=timeout
+        )
         return res.text, res.status_code
 
-    def rest_get_stream(self, url, session=None, verify=True, cert=None):
+    def rest_get_stream(self, url, session=None, verify=True, cert=None, timeout=None):
         """
         Perform a chunked GET request to url with requests.session
         This is specifically to download files.
         """
-        res = session.get(url, stream=True, verify=verify, cert=cert)
-        return res.raw, res.status_code
+        response = session.get(
+            url, stream=True, verify=verify, cert=cert, timeout=timeout
+        )
+        return response
 
     def get_stat_json(self, pathobj):
         """
@@ -632,7 +730,11 @@ class _ArtifactoryAccessor(pathlib._Accessor):
         )
 
         text, code = self.rest_get(
-            url, session=pathobj.session, verify=pathobj.verify, cert=pathobj.cert
+            url,
+            session=pathobj.session,
+            verify=pathobj.verify,
+            cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
         if code == 404 and ("Unable to find item" in text or "Not Found" in text):
             raise OSError(2, "No such file or directory: '%s'" % url)
@@ -645,19 +747,6 @@ class _ArtifactoryAccessor(pathlib._Accessor):
         """
         Request remote file/directory status info
         Returns an object of class ArtifactoryFileStat.
-
-        The following fields are available:
-          ctime -- file creation time
-          mtime -- file modification time
-          created_by -- original uploader
-          modified_by -- last user modifying the file
-          mime_type -- MIME type of the file
-          size -- file size
-          sha1 -- SHA1 digest of the file
-          sha256 -- SHA256 digest of the file
-          md5 -- MD5 digest of the file
-          is_dir -- 'True' if path is a directory
-          children -- list of children names
         """
         jsn = self.get_stat_json(pathobj)
 
@@ -737,7 +826,11 @@ class _ArtifactoryAccessor(pathlib._Accessor):
 
         url = str(pathobj) + "/"
         text, code = self.rest_put(
-            url, session=pathobj.session, verify=pathobj.verify, cert=pathobj.cert
+            url,
+            session=pathobj.session,
+            verify=pathobj.verify,
+            cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code != 201:
@@ -772,7 +865,11 @@ class _ArtifactoryAccessor(pathlib._Accessor):
 
         url = str(pathobj)
         text, code = self.rest_del(
-            url, session=pathobj.session, verify=pathobj.verify, cert=pathobj.cert
+            url,
+            session=pathobj.session,
+            verify=pathobj.verify,
+            cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code not in (200, 202, 204):
@@ -790,7 +887,11 @@ class _ArtifactoryAccessor(pathlib._Accessor):
 
         url = str(pathobj)
         text, code = self.rest_put(
-            url, session=pathobj.session, verify=pathobj.verify, cert=pathobj.cert
+            url,
+            session=pathobj.session,
+            verify=pathobj.verify,
+            cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code != 201:
@@ -828,17 +929,39 @@ class _ArtifactoryAccessor(pathlib._Accessor):
         Given the nature of HTTP streaming, this object doesn't support
         seek()
         """
-        url = str(pathobj)
-        raw, code = self.rest_get_stream(
-            url, session=pathobj.session, verify=pathobj.verify, cert=pathobj.cert
-        )
+        response = self.get_response(pathobj)
+        return response.raw
 
+    def get_response(self, pathobj):
+        """
+        :param pathobj: ArtifactoryPath object
+        :return: request response
+        """
+        url = str(pathobj)
+        response = self.rest_get_stream(
+            url,
+            session=pathobj.session,
+            verify=pathobj.verify,
+            cert=pathobj.cert,
+            timeout=pathobj.timeout,
+        )
+        code = response.status_code
         if code != 200:
             raise RuntimeError(code)
 
-        return raw
+        return response
 
-    def deploy(self, pathobj, fobj, md5=None, sha1=None, sha256=None, parameters=None):
+    def deploy(
+        self,
+        pathobj,
+        fobj,
+        md5=None,
+        sha1=None,
+        sha256=None,
+        parameters=None,
+        explode_archive=None,
+        explode_archive_atomic=None,
+    ):
         """
         Uploads a given file-like object
         HTTP chunked encoding will be attempted
@@ -859,6 +982,10 @@ class _ArtifactoryAccessor(pathlib._Accessor):
             headers["X-Checksum-Sha1"] = sha1
         if sha256:
             headers["X-Checksum-Sha256"] = sha256
+        if explode_archive:
+            headers["X-Explode-Archive"] = "true"
+        if explode_archive_atomic:
+            headers["X-Explode-Archive-Atomic"] = "true"
 
         text, code = self.rest_put_stream(
             url,
@@ -867,6 +994,7 @@ class _ArtifactoryAccessor(pathlib._Accessor):
             session=pathobj.session,
             verify=pathobj.verify,
             cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code not in (200, 201):
@@ -890,13 +1018,18 @@ class _ArtifactoryAccessor(pathlib._Accessor):
         }
 
         text, code = self.rest_post(
-            url, params=params, session=src.session, verify=src.verify, cert=src.cert
+            url,
+            params=params,
+            session=src.session,
+            verify=src.verify,
+            cert=src.cert,
+            timeout=src.timeout,
         )
 
         if code not in (200, 201):
             raise RuntimeError(text)
 
-    def move(self, src, dst):
+    def move(self, src, dst, suppress_layouts=False):
         """
         Move artifact from src to dst
         """
@@ -908,10 +1041,18 @@ class _ArtifactoryAccessor(pathlib._Accessor):
             ]
         )
 
-        params = {"to": str(dst.relative_to(dst.drive)).rstrip("/")}
+        params = {
+            "to": str(dst.relative_to(dst.drive)).rstrip("/"),
+            "suppressLayouts": int(suppress_layouts),
+        }
 
         text, code = self.rest_post(
-            url, params=params, session=src.session, verify=src.verify, cert=src.cert
+            url,
+            params=params,
+            session=src.session,
+            verify=src.verify,
+            cert=src.cert,
+            timeout=src.timeout,
         )
 
         if code not in (200, 201):
@@ -937,6 +1078,7 @@ class _ArtifactoryAccessor(pathlib._Accessor):
             session=pathobj.session,
             verify=pathobj.verify,
             cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code == 404 and ("Unable to find item" in text or "Not Found" in text):
@@ -971,6 +1113,7 @@ class _ArtifactoryAccessor(pathlib._Accessor):
             session=pathobj.session,
             verify=pathobj.verify,
             cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code == 404 and ("Unable to find item" in text or "Not Found" in text):
@@ -1004,6 +1147,7 @@ class _ArtifactoryAccessor(pathlib._Accessor):
             session=pathobj.session,
             verify=pathobj.verify,
             cert=pathobj.cert,
+            timeout=pathobj.timeout,
         )
 
         if code == 404 and ("Unable to find item" in text or "Not Found" in text):
@@ -1014,6 +1158,25 @@ class _ArtifactoryAccessor(pathlib._Accessor):
     def scandir(self, pathobj):
         return _ScandirIter((pathobj.joinpath(x) for x in self.listdir(pathobj)))
 
+    def writeto(self, pathobj, file, chunk_size, progress_func):
+        """
+        Downloads large file in chunks and prints progress
+        :param pathobj: path like object
+        :param file: IO object
+        :param chunk_size: chunk size in bytes, recommend. eg 1024*1024 is 1Mb
+        :param progress_func: Provide custom function to print out or suppress print by setting to None
+        :return: None
+        """
+
+        response = self.get_response(pathobj)
+        file_size = int(response.headers.get("Content-Length", 0))
+        bytes_read = 0
+        for chunk in response.iter_content(chunk_size=chunk_size):
+            bytes_read += len(chunk)
+            if callable(progress_func):
+                progress_func(bytes_read, file_size)
+            file.write(chunk)
+
 
 _artifactory_accessor = _ArtifactoryAccessor()
 
@@ -1060,7 +1223,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
 
     # Pathlib limits what members can be present in 'Path' class,
     # so authentication information has to be added via __slots__
-    __slots__ = ("auth", "verify", "cert", "session")
+    __slots__ = ("auth", "verify", "cert", "session", "timeout")
 
     def __new__(cls, *args, **kwargs):
         """
@@ -1090,6 +1253,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
 
         obj.cert = kwargs.get("cert")
         obj.session = kwargs.get("session")
+        obj.timeout = kwargs.get("timeout")
 
         if obj.cert is None and cfg_entry:
             obj.cert = cfg_entry["cert"]
@@ -1106,6 +1270,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
             obj.session.auth = obj.auth
             obj.session.cert = obj.cert
             obj.session.verify = obj.verify
+            obj.session.timeout = obj.timeout
 
         return obj
 
@@ -1115,6 +1280,16 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
 
         super(ArtifactoryPath, self)._init(*args, **kwargs)
 
+    @property
+    def top(self):
+        obj = ArtifactoryPath(self.drive)
+        obj.auth = self.auth
+        obj.verify = self.verify
+        obj.cert = self.cert
+        obj.session = self.session
+        obj.timeout = self.timeout
+        return obj
+
     @property
     def parent(self):
         """
@@ -1125,8 +1300,47 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
+    @property
+    def replication_status(self):
+        """
+        Get status of the repo replication
+        :return:
+            (dict): full response, where keys:
+                {status}= never_run|incomplete(running or interrupted)|error|warn|ok|inconsistent
+                {time}= time in ISO8601 format (yyyy-MM-dd'T'HH:mm:ss.SSSZ), or null if never completed
+        """
+        replication_url = self.drive + "/api/replication/" + self.repo
+        replication_obj = self.joinpath(replication_url)
+        resp = self._accessor.get_response(replication_obj).json()
+
+        return resp
+
+    def stat(self, pathobj=None):
+        """
+        Request remote file/directory status info
+        Returns an object of class ArtifactoryFileStat.
+        :param pathobj: (Optional) path like object for which to get stats.
+            if None is provided then applied to ArtifactoryPath itself
+
+        The following fields are available:
+          ctime -- file creation time
+          mtime -- file modification time
+          created_by -- original uploader
+          modified_by -- last user modifying the file
+          mime_type -- MIME type of the file
+          size -- file size
+          sha1 -- SHA1 digest of the file
+          sha256 -- SHA256 digest of the file
+          md5 -- MD5 digest of the file
+          is_dir -- 'True' if path is a directory
+          children -- list of children names
+        """
+        pathobj = pathobj or self
+        return self._accessor.stat(pathobj=pathobj)
+
     def with_name(self, name):
         """
         Return a new path with the file name changed.
@@ -1136,6 +1350,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     def with_suffix(self, suffix):
@@ -1147,8 +1362,37 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
+    def archive(self, archive_type="zip", check_sum=False):
+        """
+        Convert URL to the new link to download specified folder as archive according to REST API.
+        Requires Enable Folder Download to be set in artifactory.
+        :param: archive_type (str): one of possible archive types (supports zip/tar/tar.gz/tgz)
+        :param: check_sum (bool): defines of check sum is required along with download
+        :return: raw object for download
+        """
+        if self.is_file():
+            raise OSError("Only folders could be archived")
+
+        if archive_type not in ["zip", "tar", "tar.gz", "tgz"]:
+            raise NotImplementedError(archive_type + " is not support by current API")
+
+        archive_url = (
+            self.drive
+            + "/api/archive/download/"
+            + self.repo
+            + self.path_in_repo
+            + "?archiveType="
+            + archive_type
+        )
+
+        if check_sum:
+            archive_url += "&includeChecksumFiles=true"
+
+        return self.joinpath(archive_url)
+
     def relative_to(self, *other):
         """
         Return the relative path to another path identified by the passed
@@ -1160,6 +1404,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     def joinpath(self, *args):
@@ -1174,6 +1419,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     def __truediv__(self, key):
@@ -1185,6 +1431,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     def __rtruediv__(self, key):
@@ -1196,6 +1443,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     if sys.version_info < (3,):
@@ -1208,6 +1456,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     def _make_child_relpath(self, args):
@@ -1216,6 +1465,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         obj.verify = self.verify
         obj.cert = self.cert
         obj.session = self.session
+        obj.timeout = self.timeout
         return obj
 
     def __iter__(self):
@@ -1243,29 +1493,13 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
 
     def download_folder_archive(self, archive_type="zip", check_sum=False):
         """
-            Convert URL to the new link to download specified folder as archive according to REST API.
-            Requires Enable Folder Download to be set in artifactory.
-            :param: archive_type (str): one of possible archive types (supports zip/tar/tar.gz/tgz)
-            :param: check_sum (bool): defines of check sum is required along with download
-            :return: raw object for download
+        Convert URL to the new link to download specified folder as archive according to REST API.
+        Requires Enable Folder Download to be set in artifactory.
+        :param: archive_type (str): one of possible archive types (supports zip/tar/tar.gz/tgz)
+        :param: check_sum (bool): defines of check sum is required along with download
+        :return: raw object for download
         """
-        if archive_type not in ["zip", "tar", "tar.gz", "tgz"]:
-            raise NotImplementedError(archive_type + " is not support by current API")
-
-        archive_url = (
-            self.drive
-            + "/api/archive/download/"
-            + self.repo
-            + self.path_in_repo
-            + "?archiveType="
-            + archive_type
-        )
-
-        if check_sum:
-            archive_url += "&includeChecksumFiles=true"
-
-        with self.joinpath(archive_url) as archive_cls:
-            return self._accessor.open(archive_cls)
+        return self._accessor.open(self.archive(archive_type, check_sum))
 
     def owner(self):
         """
@@ -1361,16 +1595,39 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         """
         raise NotImplementedError()
 
-    def deploy(self, fobj, md5=None, sha1=None, sha256=None, parameters={}):
+    def deploy(
+        self,
+        fobj,
+        md5=None,
+        sha1=None,
+        sha256=None,
+        parameters={},
+        explode_archive=None,
+        explode_archive_atomic=None,
+    ):
         """
         Upload the given file object to this path
         """
         return self._accessor.deploy(
-            self, fobj, md5=md5, sha1=sha1, sha256=sha256, parameters=parameters
+            self,
+            fobj,
+            md5=md5,
+            sha1=sha1,
+            sha256=sha256,
+            parameters=parameters,
+            explode_archive=explode_archive,
+            explode_archive_atomic=explode_archive_atomic,
         )
 
     def deploy_file(
-        self, file_name, calc_md5=True, calc_sha1=True, calc_sha256=True, parameters={}
+        self,
+        file_name,
+        calc_md5=True,
+        calc_sha1=True,
+        calc_sha256=True,
+        parameters={},
+        explode_archive=False,
+        explode_archive_atomic=False,
     ):
         """
         Upload the given file to this path
@@ -1386,7 +1643,13 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
 
         with open(file_name, "rb") as fobj:
             target.deploy(
-                fobj, md5=md5, sha1=sha1, sha256=sha256, parameters=parameters
+                fobj,
+                md5=md5,
+                sha1=sha1,
+                sha256=sha256,
+                parameters=parameters,
+                explode_archive=explode_archive,
+                explode_archive_atomic=explode_archive_atomic,
             )
 
     def deploy_deb(
@@ -1413,7 +1676,7 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
 
     def copy(self, dst, suppress_layouts=False):
         """
-        Copy artifact from this path to destinaiton.
+        Copy artifact from this path to destination.
         If files are on the same instance of artifactory, lightweight (local)
         copying will be attempted.
 
@@ -1464,9 +1727,15 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
             with self.open() as fobj:
                 dst.deploy(fobj)
 
-    def move(self, dst):
+    def move(self, dst, suppress_layouts=False):
         """
         Move artifact from this path to destinaiton.
+
+        The suppress_layouts parameter, when set to True, will allow artifacts
+        from one path to be moved directly into another path without enforcing
+        repository layouts. The default behaviour is to move the repository
+        root, but remap the [org], [module], [baseVer], etc. structure to the
+        target repository.
         """
         if self.drive.rstrip("/") != dst.drive.rstrip("/"):
             raise NotImplementedError("Moving between instances is not implemented yet")
@@ -1571,9 +1840,44 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
             verify=self.verify,
             cert=self.cert,
             session=self.session,
+            timeout=self.timeout,
         )
         return obj
 
+    def promote_docker_image(
+        self, source_repo, target_repo, docker_repo, tag, copy=False
+    ):
+        """
+        Promote Docker image from source repo to target repo
+        :param source_repo: source repository
+        :param target_repo: target repository
+        :param docker_repo: Docker repository to promote
+        :param tag: Docker tag to promote
+        :param copy (bool): whether to move the image or copy it
+        :return:
+        """
+        promote_url = "{}/api/docker/{}/v2/promote".format(
+            self.drive.rstrip("/"), source_repo
+        )
+        promote_data = {
+            "targetRepo": target_repo,
+            "dockerRepository": docker_repo,
+            "tag": tag,
+            "copy": copy,
+        }
+        r = self.session.post(promote_url, json=promote_data)
+        if (
+            r.status_code == 400
+            and "Unsupported docker" in r.text
+            or r.status_code == 403
+            and "No permission" in r.text
+            or r.status_code == 404
+            and "Unable to find" in r.text
+        ):
+            raise ArtifactoryException(r.text)
+        else:
+            r.raise_for_status()
+
     @property
     def repo(self):
         return self._root.replace("/", "")
@@ -1597,29 +1901,64 @@ class ArtifactoryPath(pathlib.Path, PureArtifactoryPath):
         return None
 
     def find_repository_local(self, name):
-        obj = RepositoryLocal(self, name, packageType=None)
+        obj = RepositoryLocal(self, name)
         if obj.read():
             return obj
         return None
 
     def find_repository_virtual(self, name):
-        obj = RepositoryVirtual(self, name, packageType=None)
+        obj = RepositoryVirtual(self, name)
         if obj.read():
             return obj
         return None
 
     def find_repository_remote(self, name):
-        obj = RepositoryRemote(self, name, packageType=None)
+        obj = RepositoryRemote(self, name)
         if obj.read():
             return obj
         return None
 
+    def find_repository(self, name):
+        try:
+            return self.find_repository_local(name)
+        except ArtifactoryException:
+            pass
+
+        try:
+            return self.find_repository_remote(name)
+        except ArtifactoryException:
+            pass
+
+        try:
+            return self.find_repository_virtual(name)
+        except ArtifactoryException:
+            pass
+
+        return None
+
     def find_permission_target(self, name):
         obj = PermissionTarget(self, name)
         if obj.read():
             return obj
         return None
 
+    def writeto(self, out, chunk_size=1024, progress_func=log_download_progress):
+        """
+        Downloads large file in chunks and and call a progress function.
+
+        :param out: file path of output file
+        :param chunk_size: chunk size in bytes, recommend. eg 1024*1024 is 1MiB
+        :param progress_func: Provide custom function to print output or suppress print by setting to None
+        :return: None
+        """
+        if isinstance(out, str) or isinstance(out, pathlib.Path):
+            context = open(out, "wb")
+        else:
+            context = nullcontext(out)
+
+        with context as file:
+            self._accessor.writeto(self, file, chunk_size, progress_func)
+
     def _get_all(self, lazy: bool, url=None, key="name", cls=None):
         """
         Create a list of objects from the given endpoint
